import os
import io
import socket
import paramiko
import logging
import warnings

from typing import Union

from . import exceptions

logger = logging.getLogger()

# Cryptopgraphy update causes warnings when making SSH connections. Silence them until paramiko release a new version
# with this fix in it: https://github.com/paramiko/paramiko/pull/1379
warnings.filterwarnings(action="ignore", module=".*paramiko.*")


def _load_from_file(private_key_path: str, password: str = None) -> paramiko.RSAKey:
    try:
        return paramiko.RSAKey.from_private_key_file(private_key_path, password)
    except paramiko.ssh_exception.SSHException as e:
        logger.error(e)
        raise exceptions.PrivateKeyError("The supplied private_key file does not contain a valid private key string.")


def _load_from_string(private_key_string: str, password: str = None) -> paramiko.RSAKey:
    try:
        return paramiko.RSAKey.from_private_key(io.StringIO(private_key_string), password)
    except paramiko.ssh_exception.SSHException as e:
        logger.error(e)
        raise exceptions.PrivateKeyError("The supplied private_key string is not valid.")


def _load_private_key(private_key: str, password: str = None) -> paramiko.RSAKey:
    if os.path.isfile(private_key):
        return _load_from_file(private_key, password)
    else:
        return _load_from_string(private_key, password)


def _get_shell(client: paramiko.SSHClient, shell_timeout: int = 5):
    shell = client.invoke_shell()
    shell.settimeout(shell_timeout)
    return shell


def _get_client(
    address: str,
    port: Union[int, str],
    username: Union[str, None] = None,
    password: Union[str, None] = None,
    private_key: Union[str, None] = None,
    passphrase: Union[str, None] = None,
    allow_agent: bool = False,
    timeout: int = 10,
    missing_host_key_policy: paramiko.MissingHostKeyPolicy = paramiko.AutoAddPolicy(),
):
    auth_args = {
        "username": username,
        "password": password,
        "pkey": _load_private_key(private_key, passphrase) if private_key else None,
        "passphrase": passphrase,
    }
    client = paramiko.SSHClient()
    client.set_missing_host_key_policy(missing_host_key_policy)
    try:
        logger.info(f"opening connection to {address}:{port}...")
        client.connect(address, port=port, allow_agent=allow_agent, look_for_keys=False, timeout=timeout, **auth_args)
        return client
    except paramiko.ssh_exception.NoValidConnectionsError as e:
        #  http://docs.paramiko.org/en/2.4/api/ssh_exception.html#paramiko.ssh_exception.NoValidConnectionsError
        errors = ", ".join([str(v) for v in list(e.errors.values())])
        raise exceptions.ConnectionFailed(f"Failed to connect to {address}:{port} - {errors}.")
    except socket.timeout:
        raise exceptions.ConnectionFailed(f"Connection to {address}:{port} timed out.")
    except paramiko.ssh_exception.AuthenticationException:
        raise exceptions.AuthenticationFailed(f"Authentication failed connecting to {address}:{port}.")
    except (socket.error, paramiko.ssh_exception.SSHException) as e:
        raise exceptions.ConnectionFailed(f"Connection to {address}:{port} failed. {e}")
