"""
Users

This module contains actions for auditing & managing users on an Ubuntu server.

The following files will be used to audit user accounts & UID ranges:

- **/etc/passwd**
- **/etc/login.defs**
"""
import re
import logging

from typing import List, Dict, Union

from ppa import ui
from ssh_plugin_driver.client import run_list

from ._wrappers import ssh_connection, PROMPT_TIMEOUT, CONNECTION_TIMEOUT
from .records import UbuntuServer, User, UserData
from . import exceptions


logger = logging.getLogger(__name__)


USERS_TABLE_HEADER = ["Username", "UID", "GID", "Description", "Home Directory", "Login Shell"]
USERS_TABLE_FIELDS = ["username", "uid", "gid", "description", "home_directory", "login_shell"]
PASSWORD_FILE_COLUMNS = [0, 2, 3, 4, 5, 6]


def _check_exit_code(exit_code, command_output):
    try:
        exit_code = int(exit_code)
    except ValueError as e:
        logger.error(f"command output: {command_output}")
        logger.error(f"exit code: {exit_code}")
        message = "Failed to check the command exit code as it isn't a number. Please check the logs for more information."
        ui.output_error(message)
        raise e

    if exit_code != 0:
        message = f"Command failed with exit code {exit_code}. Output: '{command_output}'."
        ui.output_error(message)
        raise exceptions.CommandError(message)


def _parse_passwd_file(contents: str) -> List[User]:
    # https://en.wikipedia.org/wiki/Passwd#Password_file
    try:
        return [
            UserData(**{field: user[index] or None for field, index in zip(USERS_TABLE_FIELDS, PASSWORD_FILE_COLUMNS)}).__dict__
            for user in [line.strip().split(":") for line in contents.split("\n") if line]
        ]
    except Exception as e:
        logger.error("failed to parse the /etc/passwd file")
        logger.error(f"passwd file contents: {contents}")
        ui.output_error("Failed to parse /etc/passwd file. Please view the logs for more information.")
        raise e


def _parse_uid_range(output: str) -> Dict[str, int]:
    # UID_MIN			 1000
    # UID_MAX			60000
    regex = re.compile(r"UID_MIN\s+(?P<min>\d+)\sUID_MAX\s+(?P<max>\d+)")
    output = " ".join(output.split("\n"))

    if match := re.search(regex, output):
        logger.info(f"uid range: {match.groupdict()}")
        return {key: int(value) for key, value in match.groupdict().items()}

    message = "Couldn't establish UID range on the server. Please view the logs for more information."
    logger.error(f"failed to parse uid ranges from /etc/login.defs: {output}")
    ui.output_error(message)
    raise ValueError(message)


@ssh_connection
def get_regular_uid_range(ubuntu_server: UbuntuServer) -> Dict[str, int]:
    """Get the UID range for regular user accounts.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __ubuntu_server:__ an [SSHPasswordAuth](../inputs_outputs/#sshpasswordauth) or [SSHKeyAuth](../inputs_outputs/#sshkeyauth) dictionary

??? success "Output"

    A dictionary containing the following keys & values:

    - __min__: the start of the UID range
    - __max__: the end of the UID range

??? example

    ```YAML linenums="1"
    - ubuntu.users.get_regular_uid_range:
      load:
        ubuntu_server: ubuntu_server
      save: uid_range
    ```
"""
    output, exit_code = run_list(
        ubuntu_server,
        [{"command": "grep --color=never -E '^UID_MIN|^UID_MAX' /etc/login.defs"}, {"command": "echo $?"}],
        PROMPT_TIMEOUT,
        CONNECTION_TIMEOUT,
    )
    _check_exit_code(exit_code, output)
    return _parse_uid_range(output)


@ssh_connection
def get_all(ubuntu_server: UbuntuServer) -> List[User]:
    """Get all users (including system users).

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __ubuntu_server:__ an [SSHPasswordAuth](../inputs_outputs/#sshpasswordauth) or [SSHKeyAuth](../inputs_outputs/#sshkeyauth) dictionary

??? success "Output"

    A list of [User](../inputs_outputs/#user) Dictionaries.

??? example

    ```YAML linenums="1"
    - ubuntu.users.get_all:
      load:
        ubuntu_server: ubuntu_server
      save: all_users
    ```
"""
    output, exit_code = run_list(
        ubuntu_server, [{"command": "cat /etc/passwd"}, {"command": "echo $?"}], PROMPT_TIMEOUT, CONNECTION_TIMEOUT
    )
    _check_exit_code(exit_code, output)
    return _parse_passwd_file(output)


@ssh_connection
def get_regular(ubuntu_server: UbuntuServer) -> List[User]:
    """Get all _regular_ user accounts.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __ubuntu_server:__ an [SSHPasswordAuth](../inputs_outputs/#sshpasswordauth) or [SSHKeyAuth](../inputs_outputs/#sshkeyauth) dictionary

??? success "Output"

    A list of [User](../inputs_outputs/#user) Dictionaries.

??? example

    ```YAML linenums="1"
    - ubuntu.users.get_regular:
      load:
        ubuntu_server: ubuntu_server
      save: regular_users
    ```
"""
    uid_range = get_regular_uid_range(ubuntu_server)
    return list(filter(lambda u: u["uid"] >= uid_range["min"] and u["uid"] <= uid_range["max"], get_all(ubuntu_server)))


@ssh_connection
def get_system(ubuntu_server: UbuntuServer) -> List[User]:
    """Get all _system_ user accounts.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __ubuntu_server:__ an [SSHPasswordAuth](../inputs_outputs/#sshpasswordauth) or [SSHKeyAuth](../inputs_outputs/#sshkeyauth) dictionary

??? success "Output"

    A list of [User](../inputs_outputs/#user) Dictionaries.

??? example

    ```YAML linenums="1"
    - ubuntu.users.get_system:
      load:
        ubuntu_server: ubuntu_server
      save: system_users
    ```
"""
    uid_range = get_regular_uid_range(ubuntu_server)
    return list(filter(lambda u: u["uid"] < uid_range["min"] or u["uid"] > uid_range["max"], get_all(ubuntu_server)))


def output_table(text: str, users: Union[User, List[User]]) -> None:
    """Display a list of users in a table.

The table will have the following columns:

- Username
- UID
- GID
- Description
- Home Directory
- Login Shell

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __users:__ a list of [User](../inputs_outputs/#user) Dictionaries to display in the table

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - ubuntu.users.get_regular:
      load:
        ubuntu_server: ubuntu_server
      save: regular_users

    - ubuntu.users.output_table:
        text: Regular User Accounts
      load:
        users: regular_users
    ```
    """
    ui.output_table(text, header=USERS_TABLE_HEADER, data=[[u.get(field) or "" for field in USERS_TABLE_FIELDS] for u in users])


def output_tabbed_table(text: str, tabs: Dict[str, List[User]]) -> None:
    """Display a users table with multiple tabs.

This action can be used to display user lists from multiple servers.

The table will have the following columns:

- Username
- UID
- GID
- Description
- Home Directory
- Login Shell

!!! question "Minimum Plugin Version: 1.1.0"

??? info "Input"

    - __text:__ the title of the table

    - __tabs:__ a dictionary where each key is a tab name & each value is a list of [User](../inputs_outputs/#user) dictionaries

??? success "Output"

    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - set:
        name: tabbed_table
        value: {}

    - ubuntu.users.get_all:
      load:
        ubuntu_server: ubuntu_server
      save: all_users

    - ppa_tools.dictionaries.insert:
        name: My Server
      load:
        value: all_users
        dictionary: tabbed_table
      save: tabbed_table

    - ubuntu.users.output_tabbed_table:
        text: All Local Users
      load:
        tabs: tabbed_table
    ```
    """
    ui.output_table(
        text,
        header=USERS_TABLE_HEADER,
        data={key: [[value.get(field) or "" for field in USERS_TABLE_FIELDS] for value in value] for key, value in tabs.items()},
    )
